C
C SPDX-License-Identifier: BSD-3-Clause
C
      LOGICAL FUNCTION  SELECT( PAR1, PAR2 )
C
C     PURPOSE
C
C     Void logical function for DGEES.
C
      DOUBLE PRECISION  PAR1, PAR2
C
      SELECT = .TRUE.
      RETURN
      END
