C
C SPDX-License-Identifier: BSD-3-Clause
C
      LOGICAL FUNCTION  ZELCTG( PAR1, PAR2 )
C
C     PURPOSE
C
C     Void logical function for ZGGES.
C
      COMPLEX*16        PAR1, PAR2
C
      ZELCTG = .TRUE.
      RETURN
      END
