C
C SPDX-License-Identifier: BSD-3-Clause
C
      LOGICAL FUNCTION  DELCTG( PAR1, PAR2, PAR3 )
C
C     PURPOSE
C
C     Void logical function for DGGES.
C
      DOUBLE PRECISION  PAR1, PAR2, PAR3
C
      DELCTG = .TRUE.
      RETURN
      END
